@@ -16,27 +16,27 @@
 //
 //
 
-#include <grpc/support/port_platform.h>
-
 #include "src/core/lib/surface/init.h"
 
-#include "absl/base/thread_annotations.h"
-
+#include <address_sorting/address_sorting.h>
 #include <grpc/fork.h>
 #include <grpc/grpc.h>
 #include <grpc/impl/channel_arg_names.h>
 #include <grpc/support/log.h>
+#include <grpc/support/port_platform.h>
 #include <grpc/support/sync.h>
 #include <grpc/support/time.h>
 
-#include "src/core/ext/filters/client_channel/backup_poller.h"
-#include "src/core/lib/config/core_configuration.h"
+#include "absl/base/thread_annotations.h"
+#include "absl/log/log.h"
+#include "absl/time/clock.h"
+#include "absl/time/time.h"
+#include "src/core/client_channel/backup_poller.h"
+#include "src/core/config/core_configuration.h"
 #include "src/core/lib/debug/trace.h"
 #include "src/core/lib/event_engine/posix_engine/timer_manager.h"
 #include "src/core/lib/experiments/config.h"
-#include "src/core/lib/gprpp/fork.h"
-#include "src/core/lib/gprpp/sync.h"
-#include "src/core/lib/gprpp/thd.h"
+#include "src/core/lib/experiments/experiments.h"
 #include "src/core/lib/iomgr/exec_ctx.h"
 #include "src/core/lib/iomgr/iomgr.h"
 #include "src/core/lib/iomgr/timer_manager.h"
@@ -44,13 +44,19 @@
 #include "src/core/lib/security/credentials/credentials.h"
 #include "src/core/lib/security/security_connector/security_connector.h"
 #include "src/core/lib/security/transport/auth_filters.h"
-#include "src/core/lib/surface/api_trace.h"
 #include "src/core/lib/surface/channel_stack_type.h"
 #include "src/core/lib/surface/init_internally.h"
+#include "src/core/util/fork.h"
+#include "src/core/util/sync.h"
+#include "src/core/util/thd.h"
 
 // Remnants of the old plugin system
 void grpc_resolver_dns_ares_init(void);
 void grpc_resolver_dns_ares_shutdown(void);
+void grpc_resolver_dns_ares_reset_dns_resolver(void);
+
+extern absl::Status AresInit();
+extern void AresShutdown();
 
 #define MAX_PLUGINS 128
 
@@ -67,25 +73,18 @@ static bool g_shutting_down ABSL_GUARDED_BY(g_init_mu) = false;
 namespace grpc_core {
 void RegisterSecurityFilters(CoreConfiguration::Builder* builder) {
   builder->channel_init()
-      ->RegisterFilter<ClientAuthFilter>(GRPC_CLIENT_SUBCHANNEL)
+      ->RegisterV2Filter<ClientAuthFilter>(GRPC_CLIENT_SUBCHANNEL)
       .IfHasChannelArg(GRPC_ARG_SECURITY_CONNECTOR);
   builder->channel_init()
-      ->RegisterFilter<ClientAuthFilter>(GRPC_CLIENT_DIRECT_CHANNEL)
+      ->RegisterV2Filter<ClientAuthFilter>(GRPC_CLIENT_DIRECT_CHANNEL)
       .IfHasChannelArg(GRPC_ARG_SECURITY_CONNECTOR);
-  if (IsV3ServerAuthFilterEnabled()) {
-    builder->channel_init()
-        ->RegisterFilter<ServerAuthFilter>(GRPC_SERVER_CHANNEL)
-        .IfHasChannelArg(GRPC_SERVER_CREDENTIALS_ARG);
-  } else {
-    builder->channel_init()
-        ->RegisterFilter<LegacyServerAuthFilter>(GRPC_SERVER_CHANNEL)
-        .IfHasChannelArg(GRPC_SERVER_CREDENTIALS_ARG);
-  }
+  builder->channel_init()
+      ->RegisterFilter<ServerAuthFilter>(GRPC_SERVER_CHANNEL)
+      .IfHasChannelArg(GRPC_SERVER_CREDENTIALS_ARG);
   builder->channel_init()
       ->RegisterFilter<GrpcServerAuthzFilter>(GRPC_SERVER_CHANNEL)
       .IfHasChannelArg(GRPC_ARG_AUTHORIZATION_POLICY_PROVIDER)
-      .After<ServerAuthFilter>()
-      .After<LegacyServerAuthFilter>();
+      .After<ServerAuthFilter>();
 }
 }  // namespace grpc_core
 
@@ -116,11 +115,22 @@ void grpc_init(void) {
       g_shutting_down_cv->SignalAll();
     }
     grpc_iomgr_init();
-    grpc_resolver_dns_ares_init();
+    if (grpc_core::IsEventEngineDnsEnabled()) {
+      address_sorting_init();
+      auto status = AresInit();
+      if (!status.ok()) {
+        VLOG(2) << "AresInit failed: " << status.message();
+      } else {
+        // TODO(yijiem): remove this once we remove the iomgr dns system.
+        grpc_resolver_dns_ares_reset_dns_resolver();
+      }
+    } else {
+      grpc_resolver_dns_ares_init();
+    }
     grpc_iomgr_start();
   }
 
-  GRPC_API_TRACE("grpc_init(void)", 0, ());
+  GRPC_TRACE_LOG(api, INFO) << "grpc_init(void)";
 }
 
 void grpc_shutdown_internal_locked(void)
@@ -129,15 +139,20 @@ void grpc_shutdown_internal_locked(void)
     grpc_core::ExecCtx exec_ctx(0);
     grpc_iomgr_shutdown_background_closure();
     grpc_timer_manager_set_threading(false);  // shutdown timer_manager thread
-    grpc_resolver_dns_ares_shutdown();
+    if (grpc_core::IsEventEngineDnsEnabled()) {
+      address_sorting_shutdown();
+      AresShutdown();
+    } else {
+      grpc_resolver_dns_ares_shutdown();
+    }
     grpc_iomgr_shutdown();
   }
   g_shutting_down = false;
   g_shutting_down_cv->SignalAll();
 }
 
-void grpc_shutdown_internal(void* /*ignored*/) {
-  GRPC_API_TRACE("grpc_shutdown_internal", 0, ());
+void grpc_shutdown_from_cleanup_thread(void* /*ignored*/) {
+  GRPC_TRACE_LOG(api, INFO) << "grpc_shutdown_from_cleanup_thread";
   grpc_core::MutexLock lock(g_init_mu);
   // We have released lock from the shutdown thread and it is possible that
   // another grpc_init has been called, and do nothing if that is the case.
@@ -145,10 +160,11 @@ void grpc_shutdown_internal(void* /*ignored*/) {
     return;
   }
   grpc_shutdown_internal_locked();
+  VLOG(2) << "grpc_shutdown from cleanup thread done";
 }
 
 void grpc_shutdown(void) {
-  GRPC_API_TRACE("grpc_shutdown(void)", 0, ());
+  GRPC_TRACE_LOG(api, INFO) << "grpc_shutdown(void)";
   grpc_core::MutexLock lock(g_init_mu);
 
   if (--g_initializations == 0) {
@@ -162,17 +178,18 @@ void grpc_shutdown(void) {
              0) &&
         grpc_core::ExecCtx::Get() == nullptr) {
       // just run clean-up when this is called on non-executor thread.
-      gpr_log(GPR_DEBUG, "grpc_shutdown starts clean-up now");
+      VLOG(2) << "grpc_shutdown starts clean-up now";
       g_shutting_down = true;
       grpc_shutdown_internal_locked();
+      VLOG(2) << "grpc_shutdown done";
     } else {
       // spawn a detached thread to do the actual clean up in case we are
       // currently in an executor thread.
-      gpr_log(GPR_DEBUG, "grpc_shutdown spawns clean-up thread");
+      VLOG(2) << "grpc_shutdown spawns clean-up thread";
       g_initializations++;
       g_shutting_down = true;
       grpc_core::Thread cleanup_thread(
-          "grpc_shutdown", grpc_shutdown_internal, nullptr, nullptr,
+          "grpc_shutdown", grpc_shutdown_from_cleanup_thread, nullptr, nullptr,
           grpc_core::Thread::Options().set_joinable(false).set_tracked(false));
       cleanup_thread.Start();
     }
@@ -180,7 +197,7 @@ void grpc_shutdown(void) {
 }
 
 void grpc_shutdown_blocking(void) {
-  GRPC_API_TRACE("grpc_shutdown_blocking(void)", 0, ());
+  GRPC_TRACE_LOG(api, INFO) << "grpc_shutdown_blocking(void)";
   grpc_core::MutexLock lock(g_init_mu);
   if (--g_initializations == 0) {
     g_shutting_down = true;
@@ -203,3 +220,20 @@ void grpc_maybe_wait_for_async_shutdown(void) {
     g_shutting_down_cv->Wait(g_init_mu);
   }
 }
+
+bool grpc_wait_for_shutdown_with_timeout(absl::Duration timeout) {
+  GRPC_TRACE_LOG(api, INFO) << "grpc_wait_for_shutdown_with_timeout()";
+  const auto started = absl::Now();
+  const auto deadline = started + timeout;
+  gpr_once_init(&g_basic_init, do_basic_init);
+  grpc_core::MutexLock lock(g_init_mu);
+  while (g_initializations != 0) {
+    if (g_shutting_down_cv->WaitWithDeadline(g_init_mu, deadline)) {
+      LOG(ERROR) << "grpc_wait_for_shutdown_with_timeout() timed out.";
+      return false;
+    }
+  }
+  GRPC_TRACE_LOG(api, INFO)
+      << "grpc_wait_for_shutdown_with_timeout() took " << absl::Now() - started;
+  return true;
+}
