@@ -16,6 +16,10 @@
 
 #include "src/core/load_balancing/ring_hash/ring_hash.h"
 
+#include <grpc/impl/channel_arg_names.h>
+#include <grpc/impl/connectivity_state.h>
+#include <grpc/support/json.h>
+#include <grpc/support/port_platform.h>
 #include <inttypes.h>
 #include <stdlib.h>
 
@@ -30,31 +34,18 @@
 #include "absl/base/attributes.h"
 #include "absl/container/inlined_vector.h"
 #include "absl/log/check.h"
+#include "absl/log/log.h"
+#include "absl/random/random.h"
 #include "absl/status/status.h"
 #include "absl/status/statusor.h"
 #include "absl/strings/str_cat.h"
 #include "absl/strings/string_view.h"
 #include "absl/types/optional.h"
-
-#include <grpc/impl/channel_arg_names.h>
-#include <grpc/impl/connectivity_state.h>
-#include <grpc/support/json.h>
-#include <grpc/support/log.h>
-#include <grpc/support/port_platform.h>
-
 #include "src/core/client_channel/client_channel_internal.h"
+#include "src/core/config/core_configuration.h"
 #include "src/core/lib/address_utils/sockaddr_utils.h"
 #include "src/core/lib/channel/channel_args.h"
-#include "src/core/lib/config/core_configuration.h"
 #include "src/core/lib/debug/trace.h"
-#include "src/core/lib/gprpp/crash.h"
-#include "src/core/lib/gprpp/debug_location.h"
-#include "src/core/lib/gprpp/orphanable.h"
-#include "src/core/lib/gprpp/ref_counted.h"
-#include "src/core/lib/gprpp/ref_counted_ptr.h"
-#include "src/core/lib/gprpp/unique_type_name.h"
-#include "src/core/lib/gprpp/work_serializer.h"
-#include "src/core/lib/gprpp/xxhash_inline.h"
 #include "src/core/lib/iomgr/closure.h"
 #include "src/core/lib/iomgr/error.h"
 #include "src/core/lib/iomgr/exec_ctx.h"
@@ -67,7 +58,17 @@
 #include "src/core/load_balancing/lb_policy_registry.h"
 #include "src/core/load_balancing/pick_first/pick_first.h"
 #include "src/core/resolver/endpoint_addresses.h"
+#include "src/core/util/crash.h"
+#include "src/core/util/debug_location.h"
+#include "src/core/util/env.h"
 #include "src/core/util/json/json.h"
+#include "src/core/util/orphanable.h"
+#include "src/core/util/ref_counted.h"
+#include "src/core/util/ref_counted_ptr.h"
+#include "src/core/util/ref_counted_string.h"
+#include "src/core/util/unique_type_name.h"
+#include "src/core/util/work_serializer.h"
+#include "src/core/util/xxhash_inline.h"
 
 namespace grpc_core {
 
@@ -76,53 +77,79 @@ UniqueTypeName RequestHashAttribute::TypeName() {
   return kFactory.Create();
 }
 
-// Helper Parser method
-
-const JsonLoaderInterface* RingHashConfig::JsonLoader(const JsonArgs&) {
-  static const auto* loader =
-      JsonObjectLoader<RingHashConfig>()
-          .OptionalField("minRingSize", &RingHashConfig::min_ring_size)
-          .OptionalField("maxRingSize", &RingHashConfig::max_ring_size)
-          .Finish();
-  return loader;
-}
-
-void RingHashConfig::JsonPostLoad(const Json&, const JsonArgs&,
-                                  ValidationErrors* errors) {
-  {
-    ValidationErrors::ScopedField field(errors, ".minRingSize");
-    if (!errors->FieldHasErrors() &&
-        (min_ring_size == 0 || min_ring_size > 8388608)) {
-      errors->AddError("must be in the range [1, 8388608]");
-    }
-  }
-  {
-    ValidationErrors::ScopedField field(errors, ".maxRingSize");
-    if (!errors->FieldHasErrors() &&
-        (max_ring_size == 0 || max_ring_size > 8388608)) {
-      errors->AddError("must be in the range [1, 8388608]");
-    }
-  }
-  if (min_ring_size > max_ring_size) {
-    errors->AddError("max_ring_size cannot be smaller than min_ring_size");
-  }
-}
-
 namespace {
 
 constexpr absl::string_view kRingHash = "ring_hash_experimental";
 
+bool XdsRingHashSetRequestHashKeyEnabled() {
+  auto value = GetEnv("GRPC_EXPERIMENTAL_RING_HASH_SET_REQUEST_HASH_KEY");
+  if (!value.has_value()) return false;
+  bool parsed_value;
+  bool parse_succeeded = gpr_parse_bool_value(value->c_str(), &parsed_value);
+  return parse_succeeded && parsed_value;
+}
+
+class RingHashJsonArgs final : public JsonArgs {
+ public:
+  bool IsEnabled(absl::string_view key) const override {
+    if (key == "request_hash_header") {
+      return XdsRingHashSetRequestHashKeyEnabled();
+    }
+    return true;
+  }
+};
+
 class RingHashLbConfig final : public LoadBalancingPolicy::Config {
  public:
-  RingHashLbConfig(size_t min_ring_size, size_t max_ring_size)
-      : min_ring_size_(min_ring_size), max_ring_size_(max_ring_size) {}
+  RingHashLbConfig() = default;
+
+  RingHashLbConfig(const RingHashLbConfig&) = delete;
+  RingHashLbConfig& operator=(const RingHashLbConfig&) = delete;
+
+  RingHashLbConfig(RingHashLbConfig&& other) = delete;
+  RingHashLbConfig& operator=(RingHashLbConfig&& other) = delete;
+
   absl::string_view name() const override { return kRingHash; }
   size_t min_ring_size() const { return min_ring_size_; }
   size_t max_ring_size() const { return max_ring_size_; }
+  absl::string_view request_hash_header() const { return request_hash_header_; }
+
+  static const JsonLoaderInterface* JsonLoader(const JsonArgs&) {
+    static const auto* loader =
+        JsonObjectLoader<RingHashLbConfig>()
+            .OptionalField("minRingSize", &RingHashLbConfig::min_ring_size_)
+            .OptionalField("maxRingSize", &RingHashLbConfig::max_ring_size_)
+            .OptionalField("requestHashHeader",
+                           &RingHashLbConfig::request_hash_header_,
+                           "request_hash_header")
+            .Finish();
+    return loader;
+  }
+
+  void JsonPostLoad(const Json&, const JsonArgs&, ValidationErrors* errors) {
+    {
+      ValidationErrors::ScopedField field(errors, ".minRingSize");
+      if (!errors->FieldHasErrors() &&
+          (min_ring_size_ == 0 || min_ring_size_ > 8388608)) {
+        errors->AddError("must be in the range [1, 8388608]");
+      }
+    }
+    {
+      ValidationErrors::ScopedField field(errors, ".maxRingSize");
+      if (!errors->FieldHasErrors() &&
+          (max_ring_size_ == 0 || max_ring_size_ > 8388608)) {
+        errors->AddError("must be in the range [1, 8388608]");
+      }
+    }
+    if (min_ring_size_ > max_ring_size_) {
+      errors->AddError("maxRingSize cannot be smaller than minRingSize");
+    }
+  }
 
  private:
-  size_t min_ring_size_;
-  size_t max_ring_size_;
+  uint64_t min_ring_size_ = 1024;
+  uint64_t max_ring_size_ = 4096;
+  std::string request_hash_header_;
 };
 
 //
@@ -219,9 +246,14 @@ class RingHash final : public LoadBalancingPolicy {
     explicit Picker(RefCountedPtr<RingHash> ring_hash)
         : ring_hash_(std::move(ring_hash)),
           ring_(ring_hash_->ring_),
-          endpoints_(ring_hash_->endpoints_.size()) {
+          endpoints_(ring_hash_->endpoints_.size()),
+          resolution_note_(ring_hash_->resolution_note_),
+          request_hash_header_(ring_hash_->request_hash_header_) {
       for (const auto& p : ring_hash_->endpoint_map_) {
         endpoints_[p.second->index()] = p.second->GetInfoForPicker();
+        if (endpoints_[p.second->index()].state == GRPC_CHANNEL_CONNECTING) {
+          has_endpoint_in_connecting_state_ = true;
+        }
       }
     }
 
@@ -262,6 +294,9 @@ class RingHash final : public LoadBalancingPolicy {
     RefCountedPtr<RingHash> ring_hash_;
     RefCountedPtr<Ring> ring_;
     std::vector<RingHashEndpoint::EndpointInfo> endpoints_;
+    bool has_endpoint_in_connecting_state_ = false;
+    std::string resolution_note_;
+    RefCountedStringValue request_hash_header_;
   };
 
   ~RingHash() override;
@@ -280,9 +315,11 @@ class RingHash final : public LoadBalancingPolicy {
   // Current endpoint list, channel args, and ring.
   EndpointAddressesList endpoints_;
   ChannelArgs args_;
+  RefCountedStringValue request_hash_header_;
   RefCountedPtr<Ring> ring_;
 
   std::map<EndpointAddressSet, OrphanablePtr<RingHashEndpoint>> endpoint_map_;
+  std::string resolution_note_;
 
   // TODO(roth): If we ever change the helper UpdateState() API to not
   // need the status reported for TRANSIENT_FAILURE state (because
@@ -299,18 +336,34 @@ class RingHash final : public LoadBalancingPolicy {
 //
 
 RingHash::PickResult RingHash::Picker::Pick(PickArgs args) {
-  auto* call_state = static_cast<ClientChannelLbCallState*>(args.call_state);
-  auto* hash_attribute = static_cast<RequestHashAttribute*>(
-      call_state->GetCallAttribute(RequestHashAttribute::TypeName()));
-  if (hash_attribute == nullptr) {
-    return PickResult::Fail(absl::InternalError("hash attribute not present"));
+  // Determine request hash.
+  bool using_random_hash = false;
+  uint64_t request_hash;
+  if (request_hash_header_.as_string_view().empty()) {
+    // Being used in xDS.  Request hash is passed in via an attribute.
+    auto* call_state = static_cast<ClientChannelLbCallState*>(args.call_state);
+    auto* hash_attribute = call_state->GetCallAttribute<RequestHashAttribute>();
+    if (hash_attribute == nullptr) {
+      return PickResult::Fail(
+          absl::InternalError("hash attribute not present"));
+    }
+    request_hash = hash_attribute->request_hash();
+  } else {
+    std::string buffer;
+    auto header_value = args.initial_metadata->Lookup(
+        request_hash_header_.as_string_view(), &buffer);
+    if (header_value.has_value()) {
+      request_hash = XXH64(header_value->data(), header_value->size(), 0);
+    } else {
+      request_hash = absl::Uniform<uint64_t>(absl::BitGen());
+      using_random_hash = true;
+    }
   }
-  uint64_t request_hash = hash_attribute->request_hash();
-  const auto& ring = ring_->ring();
   // Find the index in the ring to use for this RPC.
   // Ported from https://github.com/RJ/ketama/blob/master/libketama/ketama.c
   // (ketama_get_server) NOTE: The algorithm depends on using signed integers
   // for lowp, highp, and index. Do not change them!
+  const auto& ring = ring_->ring();
   int64_t lowp = 0;
   int64_t highp = ring.size();
   int64_t index = 0;
@@ -336,26 +389,50 @@ RingHash::PickResult RingHash::Picker::Pick(PickArgs args) {
     }
   }
   // Find the first endpoint we can use from the selected index.
-  for (size_t i = 0; i < ring.size(); ++i) {
-    const auto& entry = ring[(index + i) % ring.size()];
-    const auto& endpoint_info = endpoints_[entry.endpoint_index];
-    switch (endpoint_info.state) {
-      case GRPC_CHANNEL_READY:
+  if (!using_random_hash) {
+    for (size_t i = 0; i < ring.size(); ++i) {
+      const auto& entry = ring[(index + i) % ring.size()];
+      const auto& endpoint_info = endpoints_[entry.endpoint_index];
+      switch (endpoint_info.state) {
+        case GRPC_CHANNEL_READY:
+          return endpoint_info.picker->Pick(args);
+        case GRPC_CHANNEL_IDLE:
+          new EndpointConnectionAttempter(
+              ring_hash_.Ref(DEBUG_LOCATION, "EndpointConnectionAttempter"),
+              endpoint_info.endpoint);
+          ABSL_FALLTHROUGH_INTENDED;
+        case GRPC_CHANNEL_CONNECTING:
+          return PickResult::Queue();
+        default:
+          break;
+      }
+    }
+  } else {
+    // Using a random hash.  We will use the first READY endpoint we
+    // find, triggering at most one endpoint to attempt connecting.
+    bool requested_connection = has_endpoint_in_connecting_state_;
+    for (size_t i = 0; i < ring.size(); ++i) {
+      const auto& entry = ring[(index + i) % ring.size()];
+      const auto& endpoint_info = endpoints_[entry.endpoint_index];
+      if (endpoint_info.state == GRPC_CHANNEL_READY) {
         return endpoint_info.picker->Pick(args);
-      case GRPC_CHANNEL_IDLE:
+      }
+      if (!requested_connection && endpoint_info.state == GRPC_CHANNEL_IDLE) {
         new EndpointConnectionAttempter(
             ring_hash_.Ref(DEBUG_LOCATION, "EndpointConnectionAttempter"),
             endpoint_info.endpoint);
-        ABSL_FALLTHROUGH_INTENDED;
-      case GRPC_CHANNEL_CONNECTING:
-        return PickResult::Queue();
-      default:
-        break;
+        requested_connection = true;
+      }
     }
+    if (requested_connection) return PickResult::Queue();
   }
-  return PickResult::Fail(absl::UnavailableError(absl::StrCat(
+  std::string message = absl::StrCat(
       "ring hash cannot find a connected endpoint; first failure: ",
-      endpoints_[ring[index].endpoint_index].status.message())));
+      endpoints_[ring[index].endpoint_index].status.message());
+  if (!resolution_note_.empty()) {
+    absl::StrAppend(&message, " (", resolution_note_, ")");
+  }
+  return PickResult::Fail(absl::UnavailableError(message));
 }
 
 //
@@ -365,7 +442,7 @@ RingHash::PickResult RingHash::Picker::Pick(PickArgs args) {
 RingHash::Ring::Ring(RingHash* ring_hash, RingHashLbConfig* config) {
   // Store the weights while finding the sum.
   struct EndpointWeight {
-    std::string address;  // Key by endpoint's first address.
+    std::string hash_key;  // By default, endpoint's first address.
     // Default weight is 1 for the cases where a weight is not provided,
     // each occurrence of the address will be counted a weight value of 1.
     uint32_t weight = 1;
@@ -377,8 +454,14 @@ RingHash::Ring::Ring(RingHash* ring_hash, RingHashLbConfig* config) {
   endpoint_weights.reserve(endpoints.size());
   for (const auto& endpoint : endpoints) {
     EndpointWeight endpoint_weight;
-    endpoint_weight.address =
-        grpc_sockaddr_to_string(&endpoint.addresses().front(), false).value();
+    auto hash_key =
+        endpoint.args().GetString(GRPC_ARG_RING_HASH_ENDPOINT_HASH_KEY);
+    if (hash_key.has_value()) {
+      endpoint_weight.hash_key = std::string(*hash_key);
+    } else {
+      endpoint_weight.hash_key =
+          grpc_sockaddr_to_string(&endpoint.addresses().front(), false).value();
+    }
     // Weight should never be zero, but ignore it just in case, since
     // that value would screw up the ring-building algorithm.
     auto weight_arg = endpoint.args().GetInt(GRPC_ARG_ADDRESS_WEIGHT);
@@ -428,8 +511,8 @@ RingHash::Ring::Ring(RingHash* ring_hash, RingHashLbConfig* config) {
   uint64_t min_hashes_per_host = ring_size;
   uint64_t max_hashes_per_host = 0;
   for (size_t i = 0; i < endpoints.size(); ++i) {
-    const std::string& address_string = endpoint_weights[i].address;
-    hash_key_buffer.assign(address_string.begin(), address_string.end());
+    const std::string& hash_key = endpoint_weights[i].hash_key;
+    hash_key_buffer.assign(hash_key.begin(), hash_key.end());
     hash_key_buffer.emplace_back('_');
     auto offset_start = hash_key_buffer.end();
     target_hashes += scale * endpoint_weights[i].normalized_weight;
@@ -530,11 +613,10 @@ void RingHash::RingHashEndpoint::CreateChildPolicy() {
           "pick_first", std::move(lb_policy_args));
   if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
     const EndpointAddresses& endpoint = ring_hash_->endpoints_[index_];
-    gpr_log(GPR_INFO,
-            "[RH %p] endpoint %p (index %" PRIuPTR " of %" PRIuPTR
-            ", %s): created child policy %p",
-            ring_hash_.get(), this, index_, ring_hash_->endpoints_.size(),
-            endpoint.ToString().c_str(), child_policy_.get());
+    LOG(INFO) << "[RH " << ring_hash_.get() << "] endpoint " << this
+              << " (index " << index_ << " of " << ring_hash_->endpoints_.size()
+              << ", " << endpoint.ToString() << "): created child policy "
+              << child_policy_.get();
   }
   // Add our interested_parties pollset_set to that of the newly created
   // child policy. This will make the child policy progress upon activity on
@@ -572,16 +654,13 @@ absl::Status RingHash::RingHashEndpoint::UpdateChildPolicyLocked() {
 void RingHash::RingHashEndpoint::OnStateUpdate(
     grpc_connectivity_state new_state, const absl::Status& status,
     RefCountedPtr<SubchannelPicker> picker) {
-  if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-    gpr_log(
-        GPR_INFO,
-        "[RH %p] connectivity changed for endpoint %p (%s, child_policy=%p): "
-        "prev_state=%s new_state=%s (%s)",
-        ring_hash_.get(), this,
-        ring_hash_->endpoints_[index_].ToString().c_str(), child_policy_.get(),
-        ConnectivityStateName(connectivity_state_),
-        ConnectivityStateName(new_state), status.ToString().c_str());
-  }
+  GRPC_TRACE_LOG(ring_hash_lb, INFO)
+      << "[RH " << ring_hash_.get() << "] connectivity changed for endpoint "
+      << this << " (" << ring_hash_->endpoints_[index_].ToString()
+      << ", child_policy=" << child_policy_.get()
+      << "): prev_state=" << ConnectivityStateName(connectivity_state_)
+      << " new_state=" << ConnectivityStateName(new_state) << " (" << status
+      << ")";
   if (child_policy_ == nullptr) return;  // Already orphaned.
   // Update state.
   const bool entered_transient_failure =
@@ -600,21 +679,16 @@ void RingHash::RingHashEndpoint::OnStateUpdate(
 //
 
 RingHash::RingHash(Args args) : LoadBalancingPolicy(std::move(args)) {
-  if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-    gpr_log(GPR_INFO, "[RH %p] Created", this);
-  }
+  GRPC_TRACE_LOG(ring_hash_lb, INFO) << "[RH " << this << "] Created";
 }
 
 RingHash::~RingHash() {
-  if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-    gpr_log(GPR_INFO, "[RH %p] Destroying Ring Hash policy", this);
-  }
+  GRPC_TRACE_LOG(ring_hash_lb, INFO)
+      << "[RH " << this << "] Destroying Ring Hash policy";
 }
 
 void RingHash::ShutdownLocked() {
-  if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-    gpr_log(GPR_INFO, "[RH %p] Shutting down", this);
-  }
+  GRPC_TRACE_LOG(ring_hash_lb, INFO) << "[RH " << this << "] Shutting down";
   shutdown_ = true;
   endpoint_map_.clear();
 }
@@ -628,9 +702,7 @@ void RingHash::ResetBackoffLocked() {
 absl::Status RingHash::UpdateLocked(UpdateArgs args) {
   // Check address list.
   if (args.addresses.ok()) {
-    if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-      gpr_log(GPR_INFO, "[RH %p] received update", this);
-    }
+    GRPC_TRACE_LOG(ring_hash_lb, INFO) << "[RH " << this << "] received update";
     // De-dup endpoints, taking weight into account.
     endpoints_.clear();
     std::map<EndpointAddressSet, size_t> endpoint_indices;
@@ -644,12 +716,10 @@ absl::Status RingHash::UpdateLocked(UpdateArgs args) {
             endpoint.args().GetInt(GRPC_ARG_ADDRESS_WEIGHT).value_or(1);
         int prev_weight_arg =
             prev_endpoint.args().GetInt(GRPC_ARG_ADDRESS_WEIGHT).value_or(1);
-        if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-          gpr_log(GPR_INFO,
-                  "[RH %p] merging duplicate endpoint for %s, combined "
-                  "weight %d",
-                  this, key.ToString().c_str(), weight_arg + prev_weight_arg);
-        }
+        GRPC_TRACE_LOG(ring_hash_lb, INFO)
+            << "[RH " << this << "] merging duplicate endpoint for "
+            << key.ToString() << ", combined weight "
+            << weight_arg + prev_weight_arg;
         prev_endpoint = EndpointAddresses(
             prev_endpoint.addresses(),
             prev_endpoint.args().Set(GRPC_ARG_ADDRESS_WEIGHT,
@@ -659,19 +729,20 @@ absl::Status RingHash::UpdateLocked(UpdateArgs args) {
       }
     });
   } else {
-    if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-      gpr_log(GPR_INFO, "[RH %p] received update with addresses error: %s",
-              this, args.addresses.status().ToString().c_str());
-    }
+    GRPC_TRACE_LOG(ring_hash_lb, INFO)
+        << "[RH " << this << "] received update with addresses error: "
+        << args.addresses.status();
     // If we already have an endpoint list, then keep using the existing
     // list, but still report back that the update was not accepted.
     if (!endpoints_.empty()) return args.addresses.status();
   }
   // Save channel args.
   args_ = std::move(args.args);
+  // Save config.
+  auto* config = DownCast<RingHashLbConfig*>(args.config.get());
+  request_hash_header_ = RefCountedStringValue(config->request_hash_header());
   // Build new ring.
-  ring_ = MakeRefCounted<Ring>(
-      this, static_cast<RingHashLbConfig*>(args.config.get()));
+  ring_ = MakeRefCounted<Ring>(this, config);
   // Update endpoint map.
   std::map<EndpointAddressSet, OrphanablePtr<RingHashEndpoint>> endpoint_map;
   std::vector<std::string> errors;
@@ -693,12 +764,14 @@ absl::Status RingHash::UpdateLocked(UpdateArgs args) {
     }
   }
   endpoint_map_ = std::move(endpoint_map);
+  // Update resolution note.
+  resolution_note_ = std::move(args.resolution_note);
   // If the address list is empty, report TRANSIENT_FAILURE.
   if (endpoints_.empty()) {
-    absl::Status status =
-        args.addresses.ok() ? absl::UnavailableError(absl::StrCat(
-                                  "empty address list: ", args.resolution_note))
-                            : args.addresses.status();
+    absl::Status status = args.addresses.ok()
+                              ? absl::UnavailableError(absl::StrCat(
+                                    "empty address list: ", resolution_note_))
+                              : args.addresses.status();
     channel_control_helper()->UpdateState(
         GRPC_CHANNEL_TRANSIENT_FAILURE, status,
         MakeRefCounted<TransientFailurePicker>(status));
@@ -769,16 +842,13 @@ void RingHash::UpdateAggregatedConnectivityStateLocked(
     state = GRPC_CHANNEL_TRANSIENT_FAILURE;
     start_connection_attempt = true;
   }
-  if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-    gpr_log(GPR_INFO,
-            "[RH %p] setting connectivity state to %s (num_idle=%" PRIuPTR
-            ", num_connecting=%" PRIuPTR ", num_ready=%" PRIuPTR
-            ", num_transient_failure=%" PRIuPTR ", size=%" PRIuPTR
-            ") -- start_connection_attempt=%d",
-            this, ConnectivityStateName(state), num_idle, num_connecting,
-            num_ready, num_transient_failure, endpoints_.size(),
-            start_connection_attempt);
-  }
+  GRPC_TRACE_LOG(ring_hash_lb, INFO)
+      << "[RH " << this << "] setting connectivity state to "
+      << ConnectivityStateName(state) << " (num_idle=" << num_idle
+      << ", num_connecting=" << num_connecting << ", num_ready=" << num_ready
+      << ", num_transient_failure=" << num_transient_failure
+      << ", size=" << endpoints_.size()
+      << ") -- start_connection_attempt=" << start_connection_attempt;
   // In TRANSIENT_FAILURE, report the last reported failure.
   // Otherwise, report OK.
   if (state == GRPC_CHANNEL_TRANSIENT_FAILURE) {
@@ -847,14 +917,12 @@ void RingHash::UpdateAggregatedConnectivityStateLocked(
       auto it = endpoint_map_.find(
           EndpointAddressSet(endpoints_[first_idle_index].addresses()));
       CHECK(it != endpoint_map_.end());
-      if (GRPC_TRACE_FLAG_ENABLED(ring_hash_lb)) {
-        gpr_log(GPR_INFO,
-                "[RH %p] triggering internal connection attempt for endpoint "
-                "%p (%s) (index %" PRIuPTR " of %" PRIuPTR ")",
-                this, it->second.get(),
-                endpoints_[first_idle_index].ToString().c_str(),
-                first_idle_index, endpoints_.size());
-      }
+      GRPC_TRACE_LOG(ring_hash_lb, INFO)
+          << "[RH " << this
+          << "] triggering internal connection attempt for endpoint "
+          << it->second.get() << " (" << endpoints_[first_idle_index].ToString()
+          << ") (index " << first_idle_index << " of " << endpoints_.size()
+          << ")";
       it->second->RequestConnectionLocked();
     }
   }
@@ -875,11 +943,9 @@ class RingHashFactory final : public LoadBalancingPolicyFactory {
 
   absl::StatusOr<RefCountedPtr<LoadBalancingPolicy::Config>>
   ParseLoadBalancingConfig(const Json& json) const override {
-    auto config = LoadFromJson<RingHashConfig>(
-        json, JsonArgs(), "errors validating ring_hash LB policy config");
-    if (!config.ok()) return config.status();
-    return MakeRefCounted<RingHashLbConfig>(config->min_ring_size,
-                                            config->max_ring_size);
+    return LoadFromJson<RefCountedPtr<RingHashLbConfig>>(
+        json, RingHashJsonArgs(),
+        "errors validating ring_hash LB policy config");
   }
 };
 
