@@ -17,19 +17,17 @@
 #ifndef GRPC_SRC_CORE_RESOLVER_XDS_XDS_DEPENDENCY_MANAGER_H
 #define GRPC_SRC_CORE_RESOLVER_XDS_XDS_DEPENDENCY_MANAGER_H
 
-#include <grpc/support/port_platform.h>
-
 #include "absl/container/flat_hash_map.h"
 #include "absl/container/flat_hash_set.h"
 #include "absl/strings/string_view.h"
-
-#include "src/core/ext/xds/xds_client_grpc.h"
-#include "src/core/ext/xds/xds_cluster.h"
-#include "src/core/ext/xds/xds_endpoint.h"
-#include "src/core/ext/xds/xds_listener.h"
-#include "src/core/ext/xds/xds_route_config.h"
-#include "src/core/lib/gprpp/ref_counted.h"
 #include "src/core/resolver/resolver.h"
+#include "src/core/resolver/xds/xds_config.h"
+#include "src/core/util/ref_counted.h"
+#include "src/core/xds/grpc/xds_client_grpc.h"
+#include "src/core/xds/grpc/xds_cluster.h"
+#include "src/core/xds/grpc/xds_endpoint.h"
+#include "src/core/xds/grpc/xds_listener.h"
+#include "src/core/xds/grpc/xds_route_config.h"
 
 namespace grpc_core {
 
@@ -38,83 +36,12 @@ namespace grpc_core {
 class XdsDependencyManager final : public RefCounted<XdsDependencyManager>,
                                    public Orphanable {
  public:
-  struct XdsConfig : public RefCounted<XdsConfig> {
-    // Listener resource.  Always non-null.
-    std::shared_ptr<const XdsListenerResource> listener;
-    // RouteConfig resource.  Will be populated even if RouteConfig is
-    // inlined into the Listener resource.
-    std::shared_ptr<const XdsRouteConfigResource> route_config;
-    // Virtual host.  Points into route_config.  Will always be non-null.
-    const XdsRouteConfigResource::VirtualHost* virtual_host;
-
-    struct ClusterConfig {
-      // Cluster resource.  Always non-null.
-      std::shared_ptr<const XdsClusterResource> cluster;
-      // Endpoint info for EDS and LOGICAL_DNS clusters.  If there was an
-      // error, endpoints will be null and resolution_note will be set.
-      struct EndpointConfig {
-        std::shared_ptr<const XdsEndpointResource> endpoints;
-        std::string resolution_note;
-
-        EndpointConfig(std::shared_ptr<const XdsEndpointResource> endpoints,
-                       std::string resolution_note)
-            : endpoints(std::move(endpoints)),
-              resolution_note(std::move(resolution_note)) {}
-        bool operator==(const EndpointConfig& other) const {
-          return endpoints == other.endpoints &&
-                 resolution_note == other.resolution_note;
-        }
-      };
-      // The list of leaf clusters for an aggregate cluster.
-      struct AggregateConfig {
-        std::vector<absl::string_view> leaf_clusters;
-
-        explicit AggregateConfig(std::vector<absl::string_view> leaf_clusters)
-            : leaf_clusters(std::move(leaf_clusters)) {}
-        bool operator==(const AggregateConfig& other) const {
-          return leaf_clusters == other.leaf_clusters;
-        }
-      };
-      absl::variant<EndpointConfig, AggregateConfig> children;
-
-      // Ctor for leaf clusters.
-      ClusterConfig(std::shared_ptr<const XdsClusterResource> cluster,
-                    std::shared_ptr<const XdsEndpointResource> endpoints,
-                    std::string resolution_note);
-      // Ctor for aggregate clusters.
-      ClusterConfig(std::shared_ptr<const XdsClusterResource> cluster,
-                    std::vector<absl::string_view> leaf_clusters);
-
-      bool operator==(const ClusterConfig& other) const {
-        return cluster == other.cluster && children == other.children;
-      }
-    };
-    // Cluster map.  A cluster will have a non-OK status if either
-    // (a) there was an error and we did not already have a valid
-    // resource or (b) the resource does not exist.
-    absl::flat_hash_map<std::string, absl::StatusOr<ClusterConfig>> clusters;
-
-    std::string ToString() const;
-
-    static absl::string_view ChannelArgName() {
-      return GRPC_ARG_NO_SUBCHANNEL_PREFIX "xds_config";
-    }
-    static int ChannelArgsCompare(const XdsConfig* a, const XdsConfig* b) {
-      return QsortCompare(a, b);
-    }
-    static constexpr bool ChannelArgUseConstPtr() { return true; }
-  };
-
   class Watcher {
    public:
     virtual ~Watcher() = default;
 
-    virtual void OnUpdate(RefCountedPtr<const XdsConfig> config) = 0;
-
-    // These methods are invoked when there is an error or
-    // does-not-exist on LDS or RDS only.
-    virtual void OnError(absl::string_view context, absl::Status status) = 0;
-    virtual void OnResourceDoesNotExist(std::string context) = 0;
+    virtual void OnUpdate(
+        absl::StatusOr<RefCountedPtr<const XdsConfig>> config) = 0;
   };
 
   class ClusterSubscription final : public DualRefCounted<ClusterSubscription> {
@@ -149,6 +76,10 @@ class XdsDependencyManager final : public RefCounted<XdsDependencyManager>,
   RefCountedPtr<ClusterSubscription> GetClusterSubscription(
       absl::string_view cluster_name);
 
+  void RequestReresolution();
+
+  void ResetBackoff();
+
   static absl::string_view ChannelArgName() {
     return GRPC_ARG_NO_SUBCHANNEL_PREFIX "xds_dependency_manager";
   }
@@ -171,6 +102,8 @@ class XdsDependencyManager final : public RefCounted<XdsDependencyManager>,
     ClusterWatcher* watcher = nullptr;
     // Most recent update obtained from this watcher.
     absl::StatusOr<std::shared_ptr<const XdsClusterResource>> update = nullptr;
+    // Ambient error.
+    std::string resolution_note;
   };
 
   struct EndpointConfig {
@@ -195,22 +128,25 @@ class XdsDependencyManager final : public RefCounted<XdsDependencyManager>,
   };
 
   // Event handlers.
-  void OnListenerUpdate(std::shared_ptr<const XdsListenerResource> listener);
+  void OnListenerUpdate(
+      absl::StatusOr<std::shared_ptr<const XdsListenerResource>> listener);
+  void OnListenerAmbientError(absl::Status status);
+
   void OnRouteConfigUpdate(
       const std::string& name,
-      std::shared_ptr<const XdsRouteConfigResource> route_config);
-  void OnError(std::string context, absl::Status status);
-  void OnResourceDoesNotExist(std::string context);
+      absl::StatusOr<std::shared_ptr<const XdsRouteConfigResource>>
+          route_config);
+  void OnRouteConfigAmbientError(std::string name, absl::Status status);
 
-  void OnClusterUpdate(const std::string& name,
-                       std::shared_ptr<const XdsClusterResource> cluster);
-  void OnClusterError(const std::string& name, absl::Status status);
-  void OnClusterDoesNotExist(const std::string& name);
+  void OnClusterUpdate(
+      const std::string& name,
+      absl::StatusOr<std::shared_ptr<const XdsClusterResource>> cluster);
+  void OnClusterAmbientError(const std::string& name, absl::Status status);
 
-  void OnEndpointUpdate(const std::string& name,
-                        std::shared_ptr<const XdsEndpointResource> endpoint);
-  void OnEndpointError(const std::string& name, absl::Status status);
-  void OnEndpointDoesNotExist(const std::string& name);
+  void OnEndpointUpdate(
+      const std::string& name,
+      absl::StatusOr<std::shared_ptr<const XdsEndpointResource>> endpoint);
+  void OnEndpointAmbientError(const std::string& name, absl::Status status);
 
   void OnDnsResult(const std::string& dns_name, Resolver::Result result);
   void PopulateDnsUpdate(const std::string& dns_name, Resolver::Result result,
@@ -242,6 +178,9 @@ class XdsDependencyManager final : public RefCounted<XdsDependencyManager>,
   // so reports an update to the watcher.
   void MaybeReportUpdate();
 
+  void ReportError(absl::string_view resource_type,
+                   absl::string_view resource_name, absl::string_view error);
+
   // Parameters passed into ctor.
   RefCountedPtr<GrpcXdsClient> xds_client_;
   std::shared_ptr<WorkSerializer> work_serializer_;
@@ -255,12 +194,14 @@ class XdsDependencyManager final : public RefCounted<XdsDependencyManager>,
   ListenerWatcher* listener_watcher_ = nullptr;
   std::shared_ptr<const XdsListenerResource> current_listener_;
   std::string route_config_name_;
+  std::string lds_resolution_note_;
 
   // RouteConfig state.
   RouteConfigWatcher* route_config_watcher_ = nullptr;
   std::shared_ptr<const XdsRouteConfigResource> current_route_config_;
   const XdsRouteConfigResource::VirtualHost* current_virtual_host_ = nullptr;
   absl::flat_hash_set<absl::string_view> clusters_from_route_config_;
+  std::string rds_resolution_note_;
 
   // Cluster state.
   absl::flat_hash_map<std::string, ClusterWatcherState> cluster_watchers_;
